import typing as t

import tinycss2
import tinycss2.ast as c2ast


Rule = c2ast.AtRule | c2ast.QualifiedRule
LineNodes = Rule | c2ast.Declaration


def strip_whitespace(content: list[c2ast.Node], beginning: bool = True, end: bool = True):
    start = 0
    stop = len(content) - 1
    if beginning:
        while start < stop:
            if not isinstance(content[start], c2ast.WhitespaceToken):
                break
            start += 1
    if end:
        while stop > start:
            if not isinstance(content[stop], c2ast.WhitespaceToken):
                break
            stop -= 1
    return content[start:stop+1]


def mk_whitespace(node: c2ast.Node, whitespace: str):
    if isinstance(node, Rule) and node.content:
        line = node.content[-1].source_line
        col = node.content[-1].source_column
    else:
        line = node.source_line
        col = node.source_column

    return c2ast.WhitespaceToken(
        line,
        col,
        whitespace
    )


def wrap_newlines(content: t.Iterable[c2ast.Node]):
    first = True
    for n in content:
        if first:
            yield mk_whitespace(n, '\n')
            first = False
        if not isinstance(n, c2ast.WhitespaceToken):
            yield n
            if isinstance(n, LineNodes):
                yield mk_whitespace(n, '\n')


def find_selector_start(index: int, content: list[c2ast.Node]):
    while index:
        n = content[index]
        if isinstance(n, c2ast.LiteralToken) and n == ';':
            index += 1
            break
        elif isinstance(n, c2ast.CurlyBracketsBlock):
            index += 1
            break
        index -= 1
    while isinstance(content[index], c2ast.WhitespaceToken):
        index += 1
    return index


def split_selector(content: list[c2ast.Node]):
    selector: list[c2ast.Node] = []
    for n in content:
        if isinstance(n, c2ast.LiteralToken) and n == ',':
            yield strip_whitespace(selector, end=False)
            selector = []
        else:
            selector.append(n)

    if selector:
        yield strip_whitespace(selector, end=False)


def merge_selectors(sel_1: list[c2ast.Node], sel_2: list[c2ast.Node]):
    sel_1 = strip_whitespace(sel_1)
    sel_2 = strip_whitespace(sel_2)
    first = sel_2[0]
    if isinstance(first, c2ast.LiteralToken) and first.value in {'&', ':'}:
        if first == '&':
            sel_2 = strip_whitespace(sel_2[1:])
            assert not isinstance(sel_2[0], c2ast.IdentToken)
    else:
        sel_1.append(mk_whitespace(sel_1[-1], ' '))

    sel_2.append(mk_whitespace(sel_2[-1], ' '))
    return [*sel_1, *sel_2]


def pump_at_rule(parent: c2ast.QualifiedRule, at_rule: c2ast.AtRule):
    new_qualified = c2ast.QualifiedRule(
        at_rule.source_line,
        at_rule.source_column,
        parent.prelude,
        []
    )
    new_at = c2ast.AtRule(
        at_rule.source_line,
        at_rule.source_column,
        at_rule.at_keyword,
        at_rule.lower_at_keyword,
        at_rule.prelude,
        [new_qualified]
    )

    for n in t.cast(list[c2ast.Node], at_rule.content):
        if isinstance(n, c2ast.QualifiedRule):
            new_at.content.extend(
                flatten_qual(
                    merge_selectors(parent.prelude, n.prelude),
                    n.content,
                    parse_declarations=False
                )
            )
        else:
            new_qualified.content.append(n)

    new_qualified.content = list(wrap_newlines(tinycss2.parse_declaration_list(new_qualified.content)))
    new_at.content = list(wrap_newlines(new_at.content))
    return new_at


def flatten_at(ar: c2ast.AtKeywordToken, sel: list[c2ast.Node], body: list[c2ast.Node]):
    rule = c2ast.AtRule(
        ar.source_line,
        ar.source_column,
        ar.value,
        ar.lower_value,
        list(sel),
        []
    )
    rule.content.extend(flatten_children(rule, body))
    rule.content = list(wrap_newlines(rule.content))
    yield rule


def flatten_qual(sel: list[c2ast.Node], body: list[c2ast.Node], parse_declarations=True):
    rule = c2ast.QualifiedRule(
        sel[0].source_line,
        sel[0].source_column,
        list(sel),
        []
    )
    children: list[Rule] = []
    for c in flatten_children(rule, body):
        if isinstance(c, c2ast.QualifiedRule):
            c.prelude = merge_selectors(sel, c.prelude)
            children.append(c)
        elif isinstance(c, c2ast.AtRule):
            children.append(pump_at_rule(rule, c))
    rule.prelude = strip_whitespace(rule.prelude, end=False)
    if parse_declarations:
        rule.content = tinycss2.parse_declaration_list(rule.content)
    rule.content = list(wrap_newlines(rule.content))
    yield rule
    yield from children


def flatten_children(parent: Rule, content: list[c2ast.Node]):
    last_safe = 0
    for i, n in enumerate(content):
        if isinstance(n, c2ast.CurlyBracketsBlock):
            start = find_selector_start(i-1, content)
            assert start >= last_safe
            parent.content.extend(strip_whitespace(content[last_safe:start]))
            sel = strip_whitespace(content[start:i], end=False)
            if isinstance(sel[0], c2ast.AtKeywordToken):
                yield from flatten_at(sel[0], sel[1:], n.content)
            else:
                for sub_sel in split_selector(sel):
                    yield from flatten_qual(sub_sel, n.content)
            last_safe = i + 1

    parent.content.extend(strip_whitespace(content[last_safe:]))


def flatten_one(node: c2ast.Node):
    if isinstance(node, c2ast.AtRule) and node.content:
        token = c2ast.AtKeywordToken(node.source_line, node.source_column, node.at_keyword)
        yield from flatten_at(token, node.prelude, node.content)
    elif isinstance(node, c2ast.QualifiedRule):
        yield from flatten_qual(node.prelude, node.content)
    elif not isinstance(node, c2ast.WhitespaceToken):
        yield node


def flatten_all(nodes: list[c2ast.Node]):
    for node in nodes:
        yield from wrap_newlines(flatten_one(node))


def process(code: str):
    return tinycss2.serialize(flatten_all(tinycss2.parse_stylesheet(code)))


if __name__ == '__main__':
    stylesheet = '''
@import foo;

.orange {
    color: #abc;
}

p {
    color: #fff;
    a {
        color: #55f;
    }
    .red {
        color: #f55;

        background-color: #f55;
    }
    :hover {
        color: #eee;
    }
    .orange, span, & .yellow {
        color: #ff5;
    }
    @media (max-width: 1000px) {
        width: 10px;
        height: 10px;
        span {
            margin: 0;
        }
    }
}

a {
    color: #fff;
    @swizzle .orange;
}

@media (min-width: 800px) and (max-width: 1000px) {
    h1 {
        font-size: 15px;
    }
    h2 {
        font-size: 12px;
        span {
            color: orange;
        }
    }
}

@when media(width >= 400px) and media(pointer: fine) {
    h1 {
        font-size: 20px;
    }
}
@else media(width < 200px) {
    h1 {
        font-style: bold;
    }
}
    '''
    def display_node(node: c2ast.Node, indent=0):
        if isinstance(node, Rule):
            print('\t'*indent, node, node.prelude)
        elif isinstance(node, c2ast.Declaration):
            print('\t'*indent, node, node.value)
        elif not isinstance(node, c2ast.WhitespaceToken):
            print('\t'*indent, node)
        if isinstance(node, Rule) and node.content:
            for child in node.content:
                display_node(child, indent+1)

    nodes = flatten_all(tinycss2.parse_stylesheet(stylesheet))
    for n in nodes:
        display_node(n)
    #print(process(stylesheet))



import re
from pathlib import Path

from .context import Context


def trim_ext_prefix(path: Path, match: re.Match[str]):
    _groups = match.groupdict()
    if 'stem' in _groups:
        return path.with_stem(_groups['stem'])
    if 'ext' in _groups:
        return Path(match.string[:match.start('ext')])
    return path


def to_dir(dest: Path, context: Context, path: Path, match: re.Match[str], ext: str | None = None):
    path = trim_ext_prefix(path, match) if ext else path

    rel = path.relative_to(
        context['input_dir']
        if path.is_relative_to(context['input_dir'])
        else context['working_dir']
    )
    new_path = dest / rel

    if ext:
        new_path = new_path.with_suffix(ext)

    return new_path


def to_output(context: Context, path: Path, match: re.Match[str], ext: str | None = None):
    return to_dir(context['output_dir'], context, path, match, ext)


def to_working(context: Context, path: Path, match: re.Match[str], ext: str | None = None):
    return to_dir(context['working_dir'], context, path, match, ext)



from __future__ import annotations

import re
import typing as t
from pathlib import Path

from .. import helpers
from ..step import Step
from .parser import process
if t.TYPE_CHECKING:
    from ..context import Context


def anchovy_css_step(path: Path, match: re.Match[str], context: Context):
    processed = process(path.read_text())
    target_path = helpers.to_output(context, path, match, '.css')
    target_path.parent.mkdir(parents=True, exist_ok=True)
    target_path.write_text(processed)
    yield target_path



from __future__ import annotations

import contextlib
import functools
import re
import shutil
import tempfile
import typing as t
from pathlib import Path

try:
    import rich.progress as _rich_progress
except ImportError:
    _rich_progress = None
try:
    import tqdm as _tqdm
except ImportError:
    _tqdm = None

from .step import Step


T = t.TypeVar('T')
StepFunc = t.Callable[[Path, re.Match[str], 'Context'], t.Iterable[Path]]
UnboundStep = t.Union[StepFunc, 'Step', None]
Rule = tuple[str, UnboundStep]
BoundStep = t.Callable[[Path, re.Match[str]], t.Iterable[Path]]
BuildSettingsKey = t.Literal['input_dir', 'output_dir', 'working_dir', 'purge_dirs']

class InputBuildSettings(t.TypedDict, total=False):
    input_dir: Path
    output_dir: Path
    working_dir: Path | None
    purge_dirs: bool

class BuildSettings(t.TypedDict):
    input_dir: Path
    output_dir: Path
    working_dir: Path
    purge_dirs: bool


def _progress(iterable: t.Iterable[T], desc: str) -> t.Iterable[T]:
    if _rich_progress:
        yield from _rich_progress.track(iterable, desc)
    elif _tqdm is not None:
        yield from _tqdm.tqdm(iterable, desc)
    else:
        print(desc)
        yield from iterable


def _rm_children(path: Path):
    if not path.exists():
        return
    for child in path.iterdir():
        if child.is_dir():
            shutil.rmtree(child)
        else:
            child.unlink()


class Context:
    def __init__(self,
                 settings: BuildSettings,
                 rules: list[Rule]):
        self.settings = settings
        self.rules = [(re.compile(r), self.bind(f)) for r, f in rules]

    @t.overload
    def __getitem__(self, key: t.Literal['input_dir', 'output_dir', 'working_dir']) -> Path: ...
    @t.overload
    def __getitem__(self, key: t.Literal['purge_dirs']) -> bool: ...
    def __getitem__(self, key):
        return self.settings[key]

    def bind(self, step: UnboundStep) -> BoundStep | None:
        if not step:
            return

        if isinstance(step, Step):
            step.bind(self)
            return step

        @functools.wraps(step)
        def bound(path: Path, match: re.Match[str]):
            return step(path, match, self)
        return bound

    def find_inputs(self, path: Path):
        for candidate in path.iterdir():
            if candidate.is_dir():
                yield from self.find_inputs(candidate)
            else:
                yield candidate

    def process(self, input_paths: list[Path] | None = None):
        input_paths = input_paths or list(self.find_inputs(self.settings['input_dir']))
        # We want to handle tasks in the order they're defined!
        tasks: dict[BoundStep, list[tuple[Path, re.Match[str]]]]
        tasks = {step: [] for test, step in self.rules if step}

        for path in _progress(input_paths, 'Planning...'):
            for test, step in self.rules:
                if match := test.match(str(path)):
                    # None can be used to halt further rule processing.
                    if not step:
                        break
                    tasks[step].append((path, match))

        flattened: list[tuple[BoundStep, Path, re.Match[str]]] = []
        for step, paths in tasks.items():
            flattened.extend((step, p, m) for p, m in paths)

        further_processing: list[Path] = []
        for step, path, match in _progress(flattened, 'Processing...'):
            new_paths = list(step(path, match))
            print(f'{path} ⇒ {", ".join(str(p) for p in new_paths)}')
            further_processing.extend(p for p in new_paths if p.is_relative_to(self['working_dir']))

        if further_processing:
            self.process(further_processing)

    def run(self, input_paths: list[Path] | None = None):
        if self['purge_dirs']:
            _rm_children(self['output_dir'])
            _rm_children(self['working_dir'])
        self.process(input_paths)



from __future__ import annotations

import re
import typing as t
from pathlib import Path

import commonmark
import commonmark.render.renderer
from jinja2 import Environment, FileSystemLoader, select_autoescape

from .step import Step
if t.TYPE_CHECKING:
    from .context import Context


class JinjaRenderStep(Step):
    env: Environment

    def __init__(self, env: Environment | None):
        self._temporary_env = env

    def bind(self, context: Context):
        super().bind(context)

        if self._temporary_env:
            self.env = self._temporary_env
        else:
            self.env = Environment(
                loader=FileSystemLoader(context['input_dir']),
                autoescape=select_autoescape()
            )

    def render_template(self, template_name: str, meta: dict[str, t.Any], target_path: Path):
        template = self.env.get_template(template_name)
        target_path.parent.mkdir(parents=True, exist_ok=True)
        template.stream(**meta).dump(str(target_path), encoding='utf-8')
        # Returning the path makes for easier composition.
        return target_path


class JinjaMarkdownStep(JinjaRenderStep):
    def __init__(self,
                 default_template: str | None = None,
                 md_parser: commonmark.Parser | None = None,
                 md_renderer: commonmark.render.renderer.Renderer | None = None,
                 jinja_env: Environment | None = None):
        super().__init__(jinja_env)
        self.default_template = default_template
        self.md_parser = md_parser or commonmark.Parser()
        self.md_renderer = md_renderer or commonmark.HtmlRenderer()

    def __call__(self, path: Path, match: re.Match[str]) -> t.Iterable[Path]:
        meta, content = self.extract_metadata(path.read_text())
        ast = self.md_parser.parse(content.strip())
        meta |= {'rendered_markdown': self.md_renderer.render(ast).strip()}

        target_path = self.context['output_dir'] / path.relative_to(self.context['input_dir']).with_suffix('.html')

        yield self.render_template(
            meta.get('template', self.default_template),
            meta,
            target_path
        )

    def extract_metadata(self, text: str):
        meta = {}
        lines = text.splitlines()

        i = 0
        for line in lines:
            if ':' not in line:
                break
            key, value = line.split(':', 1)
            if not key.isidentifier():
                break

            meta[key.strip()] = value.strip()
            i += 1

        return meta, '\n'.join(lines[i:])



from __future__ import annotations

import abc
import re
import shutil
import typing as t
from pathlib import Path

if t.TYPE_CHECKING:
    from .context import Context


class Step(abc.ABC):
    context: Context

    def bind(self, context: Context):
        self.context = context

    @abc.abstractmethod
    def __call__(self, path: Path, match: re.Match[str]) -> t.Iterable[Path]:
        ...


def direct_copy_step(path: Path, match: re.Match[str], context: Context):
    target_path = context['output_dir'] / path.relative_to(context['input_dir'])
    target_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(path, target_path)
    yield target_path
